/*
    LP: Laplacian Pyramid operations

    05/01/97 JSP Creation
*/

#include <assert.h>
#include <fstream.h>
#include <math.h>
#include <string.h>

#include "bitio.h"
#include "lp.h"

LP::LP (unsigned width, unsigned height, const unsigned char *image,
     unsigned levels, unsigned *bpp)
{
    // Build pyramid structure and copy initial image

    if (!width || !height || !levels || !image)
        throw "Invalid parameter";

    _width = width;
    _height = height;

    // Adjust levels if width or height is too small

    while (width >> (levels - 1) == 0 || height >> (levels - 1) == 0)
        --levels;

    _levels = levels;
    _bpp = new unsigned [levels];

    if (!_bpp)
        throw "Memory allocation error";

    for (unsigned level = 0; level < levels; level++)
        _bpp[level] = bpp[level];

    AllocateLevels ();

    // Copy image into level 0

    memcpy (_r_images[0], image, _width * _height);
}

LP::LP (const char *filename)
{
    // Read dimensions from a file, build structure,
    // and read in the pyramid

    FILE *fp = fopen (filename, "rb");

    if (!fp)
        throw "Could not open specified file for reading";

    _width = fgetc (fp) << 8;
    _width |= fgetc (fp);
    _height = fgetc (fp) << 8;
    _height |= fgetc (fp);
    _levels = fgetc (fp);

    if (!_width || !_height || !_levels || feof (fp))
        throw "Invalid LP file";

    _bpp = new unsigned [_levels];

    if (!_bpp)
        throw "Memory allocation error";

    for (int level = 0; level < (int) _levels; level++)
        _bpp[level] = fgetc (fp);

    AllocateLevels ();

    // Read pyramid from top to bottom

    BitFile bf (fp);

    for (level = _levels - 1; level >= 0; level--)
        for (unsigned index = 0; index < (_width >> level) * (_height >> level); index++)
            _r_images[level][index] = bf.Read (_bpp[level]);

    if (ferror (fp))
        throw "An error occurred while reading the input file";

    if (feof (fp))
        throw "Unexpected end of file";

    fclose (fp);
}

LP::~LP ()
{
    for (unsigned level = 0; level < _levels; level++)
    {
        delete [] _r_images[level];
        delete [] _l_images[level];
    }

    delete [] _r_images;
    delete [] _l_images;
    delete [] _bpp;
}

void LP::Compress ()
{
    // Compute the Laplacian Pyramid

    for (unsigned level = 0; level < _levels - 1; level++)
    {
        Reduce (level);
        Expand (level);
        Subtract (level);
    }
}

void LP::Quantize ()
{
    // Quantize by removing least significant bits

    for (unsigned level = 0; level < _levels; level++)
    {
        if (_bpp[level] > 8)
            throw "Bits per pixel is out of range";

        if (_bpp[level] == 8)
            continue;

        unsigned char *image = _l_images[level];

        for (unsigned index = 0; 
            index < (_width >> level) * (_height >> level); 
            index++)
        {
            // Get rid of least significant bits

            image[index] = image[index] >> (8 - _bpp[level]);
        }
    }
}

void LP::Unquantize ()
{
    // UnQuantize by adding in bits on the right hand side

    for (unsigned level = 0; level < _levels; level++)
    {
        if (_bpp[level] > 8)
            throw "Bits per pixel is out of range";

        if (_bpp[level] == 8)
            continue;

        unsigned char *image = _r_images[level];

        for (unsigned index = 0; 
            index < (_width >> level) * (_height >> level); 
            index++)
        {
            // Shift in least significant bits

            if (image[index])
            {
                // If it's less than zero, shift in 1's, otherwise,
                // shift in 0's

                image[index] = image[index] << (8 - _bpp[level]);

                if (image[index] & 0x80)
                    image[index] |= ((1 << (8 - _bpp[level])) - 1);
            }
        }
    }
}

void LP::Decompress ()
{
    // Build the original image from the Laplacian pyramid

    for (int level = _levels - 2; level >= 0; level--)
    {
        Expand (level);
        Add (level);
    }
}

const unsigned char *LP::GetRImage (unsigned level)
{
    if (level >= _levels)
        throw "Level is out of range";

    return _r_images[level];
}

const unsigned char *LP::GetLImage (unsigned level)
{
    if (level >= _levels)
        throw "Level is out of range";

    return _l_images[level];
}

void LP::GetDimensions (unsigned level, unsigned &width, unsigned &height)
{
    if (level >= _levels)
        throw "Level is out of range";

    width = _width >> level;
    height = _height >> level;
}

double LP::Entropy (int quantized)
{
    unsigned total = 0;
    unsigned counts[256];
    memset (counts, 0, 256 * sizeof (unsigned));

    // Add counts of Laplacians

    for (unsigned level = 0; level < _levels - 1; level++)
    {
        if (_bpp[level] || !quantized)
            for (unsigned index = 0; 
                index < (_width >> level) * (_height >> level); 
                index++)
            {
                ++counts[_l_images[level][index]];
                ++total;
            }
    }

    // Compute entropy

    double entropy = 0.0F;
    const double LOG_OF_2 = log (2.0F);

    for (unsigned index = 0; index < 256; index++)
        if (counts[index])
            entropy -= counts[index] * log (((double) counts[index]) / total) / LOG_OF_2;

    // Last level's entropy is computed separately because it is not a 
    // difference image and therefore has a more even distribution.

    if (_bpp[_levels - 1] || !quantized)
    {
        total = 0;
        memset (counts, 0, 256 * sizeof (unsigned));

        // Add counts of level N - 1

        for (index = 0; 
             index < (_width >> (_levels - 1)) * (_height >> (_levels - 1)); 
             index++)
        {
            ++counts[_r_images[_levels - 1][index]];
            ++total;
        }

        // Compute entropy for level N - 1

        for (index = 0; index < 256; index++)
            if (counts[index])
                entropy -= counts[index] * log (((double) counts[index]) / total) / LOG_OF_2;
    }

    return entropy / (_width * _height);
}

void LP::AllocateLevels ()
{
    _r_images = new unsigned char * [_levels];
    _l_images = new unsigned char * [_levels];

    if (!_r_images || !_l_images)
        throw "Memory allocation error";

    for (unsigned level = 0; level < _levels; level++)
    {
        _r_images[level] = new unsigned char [(_width >> level) * (_height >> level)];
        _l_images[level] = new unsigned char [(_width >> level) * (_height >> level)];

        if (!_r_images[level] || !_l_images[level])
            throw "Memory allocation error";

        memset (_r_images[level], 0, (_width >> level) * (_height >> level));
        memset (_l_images[level], 0, (_width >> level) * (_height >> level));
    }
}

void LP::Save (const char *filename)
{
    // The header of an LP file looks like this:
    // unsigned short width
    // unsigned short height
    // unsigned char levels
    // unsigned char bpp[levels]
    // unsigned char images[determined by width, height, levels, and bpp]

    FILE *fp = fopen (filename, "wb");

    if (!fp)
        throw "Cannot open specified file for output";

    fputc ((unsigned char) (_width >> 8), fp);
    fputc ((unsigned char) (_width), fp);
    fputc ((unsigned char) (_height >> 8), fp);
    fputc ((unsigned char) (_height), fp);
    fputc ((unsigned char) (_levels), fp);

    for (int level = 0; level < (int) _levels; level++)
        fputc ((unsigned char) _bpp[level], fp);

    // Save the top level Gaussian

    BitFile bf (fp);

    level = _levels - 1;

    for (unsigned index = 0; 
        index < (_width >> level) * (_height >> level); 
        index++)
        bf.Write ((unsigned char) _r_images[level][index], 
            _bpp[level]);

    // Save the Laplacians

    for (level = _levels - 2; level >= 0; level--)
        for (unsigned index = 0; index < (_width >> level) * (_height >> level); index++)
            bf.Write ((unsigned char) _l_images[level][index], _bpp[level]);

    bf.Flush ();

    if (ferror (fp))
        throw "An error occurred when writing to the output file";

    fclose (fp);
}

void LP::Reduce (unsigned level)
{
    // Reduce an image using a three by three block

    if (level >= _levels || level + 1 >= _levels)
        throw "Level is out of range";

    unsigned char *src = _r_images[level];
    unsigned char *dest = _r_images[level + 1];
    unsigned width = _width >> level;
    unsigned height = _height >> level;

    for (int y = 0; y < (int) (height - 3); y += 2)
    {
        for (int x = 0; x < (int) (width - 3); x += 2)
        {
            // Reduce argument is the top left corner
            // of the sample block

            unsigned char r = Reduce3x3 (&src[y * width + x], width);

            dest[((y / 2) * (width / 2)) + x / 2] = r;
        }
    }

    // Fix the edges

    // If the width is a multiple of two, then the reduced 
    // image has a missing column on the right hand side.

    if (!(width & 1))
    {
        for (int y = 0; y < (int) (height - 3); y += 2)
        {
            unsigned char r = Reduce2x3 (&src[y * width + width - 2], width);

            dest[((y / 2) * (width / 2)) + (width - 2) / 2] = r;
        }
    }

    // If the height is a multiple of two, then the reduced 
    // image has a missing row on the bottom.

    if (!(height & 1))
    {
        for (int x = 0; x < (int) (width - 3); x += 2)
        {
            unsigned char r = Reduce3x2 (&src[(height - 2) * width + x], width);

            dest[(((height - 2) / 2) * (width / 2)) + x / 2] = r;
        }
    }

    // If both the width and height are multiples of two, then
    // we missed the bottom right corner.

    if (!(width & 1) && !(height & 1))
    {
        unsigned char r = Reduce2x2 (&src[(height - 2) * width + (width - 2)], width);

        dest[(((height - 2) / 2) * (width / 2)) + (width - 2) / 2] = r;
    }
}

void LP::Expand (unsigned level)
{
    if (level >= _levels || level + 1 >= _levels)
        throw "Level is out of range";

    unsigned char *src = _r_images[level + 1];
    unsigned char *dest = _l_images[level];

    unsigned width = _width >> level;
    unsigned height = _height >> level;

    for (int y = 0; y < (int) (height - 3); y += 2)
    {
        for (int x = 0; x < (int) (width - 3); x += 2)
        {
            // Expand argument is the top left corner
            // of the sample block

            Expand3x3 (src[((y / 2) * (width / 2)) + x / 2], 
                &dest[y * width + x], width);
        }
    }

    // Fix the edges

    // If the width is a multiple of two, then the reduced 
    // image has a column on the right that was not expanded.

    if (!(width & 1))
    {
        for (int y = 0; y < (int) (height - 3); y += 2)
        {
            Expand2x3 (src[((y / 2) * (width / 2)) + (width - 2) / 2], 
                &dest[y * width + (width - 2)], width);
        }
    }

    // If the height is a multiple of two, then the reduced 
    // image has a row on the bottom that was not expanded.

    if (!(height & 1))
    {
        for (int x = 0; x < (int) (width - 3); x += 2)
        {
            Expand3x2 (src[(((height - 2) / 2) * (width / 2)) + x / 2], 
                &dest[(height - 2) * width + x], width);
        }
    }

    // If both the width and height are multiples of two, then
    // the bottom right corner was never expanded.

    if (!(width & 1) && !(height & 1))
    {
        Expand2x2 (src[(((height - 2) / 2) * (width / 2)) + (width - 2) / 2], 
            &dest[(height - 2) * width + (width - 2)], width);
    }

    // When expanding, the outer edges need a little more
    // fixing because they have not been interpolated enough.

    // The top row only got interpolated with half the number 
    // of pixels needed.

    for (int x = 0; x < (int) (width); x++)
        dest[x] *= 2;

    // Same with the left column.  The top left corner only got 1/4
    // the number of pixels it needed, but we're multiplying it
    // by 2 twice, so it's OK.

    for (y = 0; y < (int) (height); y++)
        dest[y * width] *= 2;

    // If the width is an odd number, the number of pixels needed
    // to interpolate the right column is 1/2.  If it's an even
    // number, then it's OK.

    if (width & 1)
        for (int y = 1; y <= (int) (height); y++)
            dest[y * width - 1] *= 2;

    // If the height is an odd number, then the bottom row needs
    // twice as many pixels to interpolate correctly.

    if (height & 1)
        for (int x = 0; x < (int) (width); x++)
            dest[(height - 1) * width + x] *= 2;

    // Notice that the NE, SE, and SW corners are OK -- they
    // were multiplied by 2 or 4 automatically when the sides
    // were being fixed.
}

void LP::Subtract (unsigned level)
{
    if (level >= _levels)
        throw "Level is out of range";

    unsigned char *src = _r_images[level];
    unsigned char *dest = _l_images[level];
    unsigned width = _width >> level;
    unsigned height = _height >> level;

    for (unsigned index = 0; index < width * height; index++)
        dest[index] = src[index] - dest[index];
}

void LP::Add (unsigned level)
{
    if (level >= _levels)
        throw "Level is out of range";

    unsigned char *src = _l_images[level];
    unsigned char *dest = _r_images[level];
    unsigned width = _width >> level;
    unsigned height = _height >> level;

    for (unsigned index = 0; index < width * height; index++)
        dest[index] += src[index];
}

unsigned char LP::Reduce3x3 (const unsigned char *block, unsigned width)
{
    // Reduce a block using a 3 by 3 equivalent weighting function

    unsigned r;

    r = *block; // First row
    r += *++block * 2;
    r += *++block;
    block += width - 2;
    r += *block * 2; // Second row
    r += *++block * 4;
    r += *++block * 2;
    block += width - 2;
    r += *block; // Third row
    r += *++block * 2;
    r += *++block;

    return r / 16;
}

unsigned char LP::Reduce2x3 (const unsigned char *block, unsigned width)
{
    // Reduce a block using a 3 by 3 equivalent weighting function

    // Reflect left column to right column

    unsigned r;

    r = *block * 2; // First row
    r += *++block * 2;
    block += width - 1;
    r += *block * 4; // Second row
    r += *++block * 4;
    block += width - 1;
    r += *block * 2; // Third row
    r += *++block * 2;

    return r / 16;
}

unsigned char LP::Reduce3x2 (const unsigned char *block, unsigned width)
{
    // Reduce a block using a 3 by 3 equivalent weighting function

    // Reflect top row to bottom row

    unsigned r;

    r = *block * 2; // First row
    r += *++block * 4;
    r += *++block * 2;
    block += width - 2;
    r += *block * 2; // Second row
    r += *++block * 4;
    r += *++block * 2;

    return r / 16;
}

unsigned char LP::Reduce2x2 (const unsigned char *block, unsigned width)
{
    // Reduce a block using a 3 by 3 equivalent weighting function

    // Reflect left column to right column
    // Reflect top row to bottom row

    unsigned r;

    r = *block * 4; // First row
    r += *++block * 4;
    block += width - 1;
    r += *block * 4; // Second row
    r += *++block * 4;

    return r / 16;
}

void LP::Expand3x3 (unsigned char val, unsigned char *block, unsigned width)
{
    *block += val / 4; // First row
    *++block += val / 2;
    *++block += val / 4;
    block += width - 2;
    *block += val / 2; // Second row
    *++block += val;
    *++block += val / 2;
    block += width - 2;
    *block += val / 4; // Third row
    *++block += val / 2;
    *++block += val / 4;
}

void LP::Expand2x3 (unsigned char val, unsigned char *block, unsigned width)
{
    *block += val / 4; // First row
    *++block += val / 2;
    block += width - 1;
    *block += val / 2; // Second row
    *++block += val;
    block += width - 1;
    *block += val / 4; // Third row
    *++block += val / 2;
}

void LP::Expand3x2 (unsigned char val, unsigned char *block, unsigned width)
{
    *block += val / 4; // First row
    *++block += val / 2;
    *++block += val / 4;
    block += width - 2;
    *block += val / 2; // Second row
    *++block += val;
    *++block += val / 2;
}

void LP::Expand2x2 (unsigned char val, unsigned char *block, unsigned width)
{
    *block += val / 16; // First row
    *++block += val / 8;
    block += width - 1;
    *block += val / 8; // Second row
    *++block += val / 4;
}
