/*
    TGA: Targa file operations

    05/01/97 JSP Creation
*/

#include <stdio.h>
#include <string.h>

#include "tga.h"

TGA::TGA ()
{
    _image = 0;
    memset (&_tga_header, 0, sizeof (_tga_header));
}

TGA::~TGA ()
{
    if (_image)
        delete [] _image;
}

void TGA::Load (const char *filename)
{
    FILE *tga_file = fopen (filename, "rb");

    if (!tga_file)
        throw "Can't open specified TGA file for reading";

    _tga_header.id_field_length = fgetc (tga_file);
    _tga_header.color_map_type = fgetc (tga_file);
    _tga_header.image_type_code = fgetc (tga_file);
    _tga_header.color_map_origin = fgetc (tga_file);
    _tga_header.color_map_origin |= fgetc (tga_file) << 8;
    _tga_header.color_map_length = fgetc (tga_file);
    _tga_header.color_map_length |= fgetc (tga_file) << 8;
    _tga_header.color_map_entry_size = fgetc (tga_file);
    _tga_header.x_origin = fgetc (tga_file);
    _tga_header.x_origin |= fgetc (tga_file) << 8;
    _tga_header.y_origin = fgetc (tga_file);
    _tga_header.y_origin |= fgetc (tga_file) << 8;
    _tga_header.width = fgetc (tga_file);
    _tga_header.width |= fgetc (tga_file) << 8;
    _tga_header.height = fgetc (tga_file);
    _tga_header.height |= fgetc (tga_file) << 8;
    _tga_header.bits_per_pixel = fgetc (tga_file);
    _tga_header.image_descriptor = fgetc (tga_file);

    for (unsigned index = 0; index < _tga_header.id_field_length; index++)
        fgetc (tga_file);

    switch (_tga_header.image_type_code)
    {
        case 0:
        throw "The image contains no data";

        case 1: // Color-mapped RGB
        case 2: // Indexed RGB
        throw "Only greyscale images are supported";
        break;

        case 3: // Greyscale
        break;

        case 9:
        case 10:
        case 11:
        throw "Compressed TGA files are not supported";

        default:
        throw "Unknown TGA file format is not supported";
    }

    if (_tga_header.color_map_type)
        throw "Only greyscale images are supported";

    // Check to make sure the header was read in OK

    if (feof (tga_file) || ferror (tga_file))
    {
        fclose (tga_file);

        throw "Invalid TGA file";
    }

    if (_tga_header.image_descriptor & INTERLEAVE_MASK)
    {
        fclose (tga_file);

        throw "Interleaved images are not supported";
    }

    if (_image)
        delete [] _image;

    _image = 0;

    if (_tga_header.bits_per_pixel != 8)
    {
        fclose (tga_file);

        throw "Only 8 bits per pixel formats are supported";
    }

    _image = new unsigned char [_tga_header.width * _tga_header.height];

    if (!_image)
    {
        fclose (tga_file);

        throw "Memory allocation error";
    }

    for (unsigned y = 0; y < _tga_header.height; y++)
    {
        unsigned y_index = y;

        if (!(_tga_header.image_descriptor & TOP_TO_BOTTOM_MASK))
            y_index = _tga_header.height - y_index - 1;

        for (unsigned x = 0; x < _tga_header.width; x++)
        {
            unsigned x_index = x;

            if ((_tga_header.image_descriptor & LEFT_TO_RIGHT_MASK))
                x_index = _tga_header.width - x_index - 1;

            unsigned index = y_index * _tga_header.width + x_index;

            _image[index] = (unsigned char) (fgetc (tga_file));
        }
    }

    // Check to make sure the image was read in OK

    if (feof (tga_file) || ferror (tga_file))
    {
        fclose (tga_file);

        throw "Invalid TGA file";
    }

    fclose (tga_file);
}

void TGA::Load (unsigned width, 
    unsigned height, 
    unsigned bits_per_pixel, 
    unsigned image_type_code,
    const void *image)
{
    switch (image_type_code)
    {
        case TGA_COLOR_MAPPED:
        case TGA_TRUE_COLOR:
        throw "Only greyscale images are supported";

        case TGA_GREY_SCALE:
        _tga_header.image_type_code = TGA_GREY_SCALE;
        break;

        default:
        throw "Unknown image type code is not supported";
    }

    _tga_header.id_field_length = 0; // No ID field
    _tga_header.color_map_type = 0; // No color map
    _tga_header.color_map_origin = 0;
    _tga_header.color_map_length = 0;
    _tga_header.color_map_entry_size = 0;
    _tga_header.x_origin = 0;
    _tga_header.y_origin = 0;
    _tga_header.width = (unsigned short) width;
    _tga_header.height = (unsigned short) height;
    _tga_header.bits_per_pixel = (unsigned char) bits_per_pixel;
    _tga_header.image_descriptor = TOP_TO_BOTTOM_MASK;

    if (_image)
        delete [] _image;

    _image = new unsigned char [_tga_header.width * _tga_header.height * BytesPerPixel ()];

    if (!_image)
        throw "Memory allocation error";

    memcpy (_image, image, _tga_header.width * _tga_header.height * BytesPerPixel ());
}

void TGA::Save (const char *filename)
{
    FILE *tga_file = fopen (filename, "wb");

    if (!tga_file)
        throw "Could not open specified TGA file for writing";

    fputc (_tga_header.id_field_length, tga_file);
    fputc (_tga_header.color_map_type, tga_file);
    fputc (_tga_header.image_type_code, tga_file);
    fputc ((char) (_tga_header.color_map_origin), tga_file);
    fputc ((char) (_tga_header.color_map_origin >> 8), tga_file);
    fputc ((char) (_tga_header.color_map_length), tga_file);
    fputc ((char) (_tga_header.color_map_length >> 8), tga_file);
    fputc (_tga_header.color_map_entry_size, tga_file);
    fputc ((char) (_tga_header.x_origin), tga_file);
    fputc ((char) (_tga_header.x_origin >> 8), tga_file);
    fputc ((char) (_tga_header.y_origin), tga_file);
    fputc ((char) (_tga_header.y_origin >> 8), tga_file);
    fputc ((char) (_tga_header.width), tga_file);
    fputc ((char) (_tga_header.width >> 8), tga_file);
    fputc ((char) (_tga_header.height), tga_file);
    fputc ((char) (_tga_header.height >> 8), tga_file);
    fputc (_tga_header.bits_per_pixel, tga_file);
    fputc (_tga_header.image_descriptor, tga_file);

    for (unsigned index = 0; index < _tga_header.width * _tga_header.height * BytesPerPixel (); index++)
        fputc (_image[index], tga_file);

    if (feof (tga_file) || ferror (tga_file))
    {
        fclose (tga_file);

        throw "Error writing to specified TGA file";
    }

    fclose (tga_file);
}
